/**
  Physics simulator demo program.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2011-01-20 (Public Domain)
*/
#include "physics/world.h"
#include "ogl/glsl.h"

/* LAME! just include library bodies here, for easy linking */
#include "physics/world.cpp" 
#include "physics/config.cpp"
#include "ogl/glsl.cpp"

/* One simulated particle */
class particle {
public:
	vec4 color; // RGBA
	vec3 pos, vel; // in world coordinates-- meters
	vec3 netforce; // in newtons (quarter-pounds)
	
	void start_step(double dt) {
		/*
		 For stability, I'm using "Leapfrog" integration:
			http://en.wikipedia.org/wiki/Leapfrog_integration
		 Basically, we just update position, *then* calculate
		 net force before moving the velocity.
		*/
		pos+=dt*vel;
		
		/* Zero out net force, so springs can add to it */
		netforce=vec3(0.0);
	}
	
	void finish_step(double dt,const vec3 &target) {
	/* Newtonian motion */
		float mass=1.0; // kilograms (for this particle)
		vec3 acc=netforce/mass; 

		acc.z+=-9.8;  /* gravity */
		
#if 0
		/* "target-seeking": must be stronger than gravity */
		vec3 toTarget=(target-pos); 
		toTarget.z=0.0;
		acc+=normalize(toTarget)*20.0;
		
		/* Hazard avoidance force */
		vec3 hazard=vec3(0.0,0.0,-1.0);
		vec3 fear=-(hazard-pos);
		fear.z=0.0;
		float fearMag=3.0-length(fear);
		if (fearMag>0) { // close enough to scare me! 
			acc+=normalize(fear)*fearMag*1.0;
		}
#endif
		
		/* "wind drag": fairly gentle */
		acc+=-1.0*vel;
		vel+=dt*acc;

		if (pos.z<0.0) { /* hit the ground */
			vel.z=0.6*fabs(vel.z);
			vel*=0.4; // ground drag
			pos.z=0.0;
		}
		if (pos.x<-2.0) { /* hit the left wall */
			vel.x=0.6*fabs(vel.x);
			pos.x=-2.0;
		}
	}
};

/* Random float between 0 and 1 */
float randf(void) {
	float f=rand();
	return f/RAND_MAX;
}

/* Random vector where x y and z vary between 0 and 1 */
vec3 randVec3(void) {
	return vec3(randf(),randf(),randf());
}


class triangle {
public:
	int v[3];
	float restlength[3]; // length in meters of rest position
};

void bad(const char *err) {
	std::cout<<"Fatal file read error:  "<<err<<"\n";
	exit(1);
}

/* A particle system */
class particlesystem : public physics::object {
	std::vector<particle> p;
	std::vector<triangle> tris;
	vec3 targetpoint;
	bool initRestState;
public:
	particlesystem(void) 
		:physics::object(0.01)  /* <- our timestep, in seconds */
	{
		initRestState=true;
		
		/* Read a Geomview ".off" 3D model.
		    http://people.sc.fsu.edu/~jburkardt/data/off/off.html
		*/
		std::ifstream f("model.off");
		std::string first; f>>first;
		if (first!="OFF") bad("expected 'OFF' as the first thing in .off file");
		
		int nvert, nface, nedge; 
		f>>nvert>>nface>>nedge;
		
		/* Read all the XYZ vertices */
		for (int i=0;i<nvert;i++) {
			float x,y,z;
			f>>x>>y>>z;
			particle np;
			np.color=1.0*randVec3(); np.color.w=0.5; /* alpha */
			np.pos=vec3(x,z,-y)+vec3(2.0,0,1.3);
			np.vel=vec3(-8.0,0.0,5.0); // (randVec3()-vec3(0.5));
			p.push_back(np);
		}
		
		/* Read all the triangles */
		for (int i=0;i<nface;i++) {
			int n; f>>n; if (n!=3) bad("can only read triangles from .off file");
			int a,b,c; f>>a>>b>>c;
			triangle t; t.v[0]=a; t.v[1]=b; t.v[2]=c;
			tris.push_back(t);
		}
		if (!f) bad("I/O error reading .off file");
		f.close();
	}
	
	void simulate(physics::library &lib) {
		float slowmo=0.1;
		float dt=timestep.dt*slowmo;
	
		// Start timesteps
		for (unsigned int i=0;i<p.size();i++)
		{
			p[i].start_step(dt);
			if (length(p[i].vel)>1000.0) {std::cout<<"Unstable at t="<<timestep.last_t<<"\n"; exit(1);}
		}
		
		// Loop over edges, add in net forces to particles
		for (unsigned int i=0;i<tris.size();i++) {
			for (int j=0;j<3;j++) {
				int a=tris[i].v[j];
				int b=tris[i].v[(j+1)%3];
				
				vec3 dist=p[a].pos-p[b].pos;
				vec3 along=normalize(dist);
				float len=length(dist);
				
				/* Newtons of force per meter of displacement:
				    this should be pretty huge, like 100000 or so.
					with millisecond timesteps, unstable above 240000
				*/
				float springconst=200000.0;
				
				if (initRestState) tris[i].restlength[j]=len-0.0001;
				
				/* Spring force is straight from Hooke's Law */
				float f=-springconst*(len-tris[i].restlength[j]);
				
				p[a].netforce+=f*along;
				p[b].netforce-=f*along;
			}
		}
		initRestState=false;
		
	
		// Finish timesteps
		for (unsigned int i=0;i<p.size();i++) {
			p[i].finish_step(dt,targetpoint);
		}
	}
	
	void draw(physics::library &lib) {
		glColor4f(0.5f,0.5f,0.5f,1.0f);
		glEnable(GL_POINT_SMOOTH);
		
		/* Red quad represents the left bumper */
		glPushMatrix();
		glTranslatef(-2.0,0,+1);
		static vec3 quadpts[4]=
			{vec3(0,-4,-4),vec3(0,+4,-4), vec3(0,+4,+4), vec3(0,-4,+4)};
		glDepthMask(GL_FALSE); // transparent, so don't mess up Z buffer
		glColor4f(1,0,0,0.3);
		glBegin(GL_QUADS); 
			for (int i=0;i<4;i++) glVertex3fv(quadpts[i]); 
		glEnd();
		glDepthMask(GL_TRUE);
		glColor3f(0,0,0);
		glBegin(GL_LINE_LOOP); 
			for (int i=0;i<4;i++) glVertex3fv(quadpts[i]); 
		glEnd();
		glPopMatrix();
		
		/* Draw each of the nodes */
		glPointSize(20.0);
		glBegin(GL_POINTS);
		for (unsigned int i=0;i<p.size();i++) {
			glColor4fv(p[i].color);
			glVertex3fv(p[i].pos);
			
			// Draw a "shadow"
			glColor4f(0.0f,0.0f,0.0f,0.05f);
			glVertex2fv(p[i].pos); /* 2fv means z==0 */
		}
		glEnd();
		
		/* Outline the triangles */
		glColor4f(1,0,0,1);
		glLineWidth(3.0);
		glBegin(GL_LINES);
		for (unsigned int i=0;i<tris.size();i++) {
			for (int j=0;j<3;j++) {
				glVertex3fv(p[tris[i].v[j]].pos);
				glVertex3fv(p[tris[i].v[(j+1)%3]].pos);
			}
		}
		glEnd();
	}
};


/* Called create a new simulation */
void physics_setup(physics::library &lib) {
	lib.world->add(new particlesystem());
	lib.background[0]=lib.background[1]=0.6;
	lib.background[2]=1.0; // light blue
}

/* Called every frame to draw everything */
void physics_draw_frame(physics::library &lib) {
	/* Point and line smoothing needs normal alpha blending */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	/* Draw opaque white ground plane */
	int groundsize=30;
	glEnable(GL_DEPTH_TEST);
	glColor3f(1,1,1);
	glBegin(GL_QUADS);
	glVertex2i(-groundsize,-groundsize); glVertex2i(+groundsize,-groundsize);
	glVertex2i(+groundsize,+groundsize); glVertex2i(-groundsize,+groundsize);
	glEnd();
	
	/* No depth while drawing ground grid */
	glDisable(GL_DEPTH_TEST);
	glColor4f(0.2,0.2,0.2,0.2);
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(1.0);
	if (camera.z>=0.0) { /* no grid if you're underneath ground */
		glBegin(GL_LINES);
		for (int grid=-groundsize;grid<=+groundsize;grid++) {
			glVertex3i(-groundsize,grid,0); glVertex3i(+groundsize,grid,0); 
			glVertex3i(grid,-groundsize,0); glVertex3i(grid,+groundsize,0); 
		}
		glEnd();
	}
	
	/* Do depth while drawing objects */
	glEnable(GL_DEPTH_TEST);
	lib.world->draw(lib);
}


